@@ -40,6 +40,9 @@ def bw_lock_add(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['targets'], args['node_workers'])
     target_nodes = remove_dummy_nodes(target_nodes)
+    if not target_nodes:
+        return
+
     pending_nodes = target_nodes[:]
     max_node_name_length = max([len(node.name) for node in target_nodes])
     lock_id = randstr(length=4).upper()
@@ -94,6 +97,9 @@ def bw_lock_remove(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['targets'], args['node_workers'])
     target_nodes = remove_dummy_nodes(target_nodes)
+    if not target_nodes:
+        return
+
     pending_nodes = target_nodes[:]
     max_node_name_length = max([len(node.name) for node in target_nodes])
     io.progress_set_total(len(pending_nodes))
@@ -150,6 +156,9 @@ def bw_lock_show(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['targets'], args['node_workers'])
     target_nodes = remove_dummy_nodes(target_nodes)
+    if not target_nodes:
+        return
+
     pending_nodes = target_nodes[:]
     locks_on_node = {}
     exit_code = 0
