@@ -171,6 +171,12 @@ bundle:my_bundle  # items in this bundle
         dest='summary',
         help=_("don't show stats summary"),
     )
+    parser_apply.add_argument(
+        "--no-skipped-items",
+        action='store_false',
+        dest='show_skipped_items',
+        help=_("don't show skipped items"),
+    )
     parser_apply.add_argument(
         "-r",
         "--resume-file",
@@ -705,7 +711,6 @@ will exit with code 47 if any matching items are locked
         "Color guide: "
         "needs in red, "
         "after in blue, "
-        "concurrency blockers in purple, "
         "before in yellow, "
         "needed_by in orange, "
         "auto in green, "
@@ -735,11 +740,12 @@ will exit with code 47 if any matching items are locked
         dest='depends_auto',
         help=_("do not show auto-generated and trigger dependencies"),
     )
+    # XXX Remove in bw 5
     parser_plot_subparsers_node.add_argument(
         "--no-depends-conc",
         action='store_false',
         dest='depends_concurrency',
-        help=_("do not show concurrency blocker dependencies"),
+        help=_("obsolete and ignored"),
     )
     parser_plot_subparsers_node.add_argument(
         "--no-depends-regular",
