@@ -15,19 +15,20 @@ class DnfPkg(Pkg):
         return ["pkg_dnf", "pkg_yum"]
 
     def pkg_all_installed(self):
-        result = self.run("dnf -d0 -e0 list installed")
-        for line in result.stdout.decode('utf-8').strip().split("\n"):
+        result = self.run("dnf list --installed")
+        # First line is a header, skip that.
+        for line in result.stdout.decode('utf-8').strip().split("\n")[1:]:
             yield "{}:{}".format(self.ITEM_TYPE_NAME, line.split()[0].split(".")[0])
 
     def pkg_install(self):
-        self.run("dnf -d0 -e0 -y install {}".format(quote(self.name)), may_fail=True)
+        self.run("dnf -y install {}".format(quote(self.name)), may_fail=True)
 
     def pkg_installed(self):
         result = self.run(
-            "dnf -d0 -e0 list installed {}".format(quote(self.name)),
+            "dnf list --installed {}".format(quote(self.name)),
             may_fail=True,
         )
         return result.return_code == 0
 
     def pkg_remove(self):
-        self.run("dnf -d0 -e0 -y remove {}".format(quote(self.name)), may_fail=True)
+        self.run("dnf -y remove {}".format(quote(self.name)), may_fail=True)
