@@ -1,5 +1,5 @@
 from contextlib import contextmanager, suppress
-from datetime import datetime
+from datetime import datetime, timedelta
 import faulthandler
 from functools import wraps
 from os import _exit, environ, getpid, kill
@@ -135,13 +135,15 @@ def page_lines(lines):
             io.stdout(line)
 
 
-def write_to_stream(stream, msg):
+def write_to_stream(stream, msg, flush=True):
     with suppress(BrokenPipeError):
         if TTY:
             stream.write(msg)
         else:
             stream.write(ansi_clean(msg))
-        stream.flush()
+
+        if flush:
+            stream.flush()
 
 
 class DrainableStdin:
@@ -157,6 +159,43 @@ class DrainableStdin:
             termios.tcflush(sys.stdin, termios.TCIFLUSH)
 
 
+class JobManager:
+    def __init__(self):
+        self._jobs = []
+
+    def add(self, msg):
+        job_id = (time(), msg)
+        self._jobs.append(job_id)
+        return job_id
+
+    def remove(self, job_id):
+        self._jobs.remove(job_id)
+
+    @property
+    def current_job(self):
+        try:
+            job_start, job_msg = self._jobs[-1]
+        except IndexError:
+            return None
+        current_time = time()
+        if current_time - job_start > 3.0:
+            # If the latest job is taking a long time, start rotating
+            # the displayed job every 3s. That way, users can see all
+            # long-running jobs currently in progress.
+            index = int(current_time / 3.0) % len(self._jobs)
+            job_start, job_msg = self._jobs[index]
+
+        elapsed = current_time - job_start
+        if elapsed > 10.0:
+            job_msg += " ({})".format(format_duration(timedelta(seconds=elapsed)))
+
+        return job_msg
+
+    @property
+    def messages(self):
+        return [job_msg for job_start, job_msg in self._jobs]
+
+
 class IOManager:
     """
     Threadsafe singleton class that handles all IO.
@@ -165,7 +204,7 @@ class IOManager:
         self._active = False
         self.debug_log_file = None
         self.debug_mode = False
-        self.jobs = []
+        self.jobs = JobManager()
         self.lock = Lock()
         self.progress = 0
         self.progress_start = None
@@ -211,7 +250,7 @@ class IOManager:
         with self.lock:
             if QUIT_EVENT.is_set():
                 sys.exit(0)
-            self._clear_last_job()
+            self._clear_last_job(flush=False)
             while True:
                 write_to_stream(STDOUT_WRITER, "\a" + question + SHOW_CURSOR)
 
@@ -266,16 +305,17 @@ class IOManager:
         if not self._active:
             return
         with self.lock:
-            self._clear_last_job()
-            self.jobs.append(msg)
+            self._clear_last_job(flush=False)
+            job_id = self.jobs.add(msg)
             self._write_current_job()
+            return job_id
 
-    def job_del(self, msg):
+    def job_del(self, job_id):
         if not self._active:
             return
         with self.lock:
-            self._clear_last_job()
-            self.jobs.remove(msg)
+            self._clear_last_job(flush=False)
+            self.jobs.remove(job_id)
             self._write_current_job()
 
     def progress_advance(self, increment=1):
@@ -295,9 +335,9 @@ class IOManager:
         if INFO_EVENT.is_set():
             INFO_EVENT.clear()
             table = []
-            if self.jobs:
-                table.append([bold(_("Running jobs")), self.jobs[0].strip()])
-                for job in self.jobs[1:]:
+            if self.jobs.messages:
+                table.append([bold(_("Running jobs")), self.jobs.messages[0].strip()])
+                for job in self.jobs.messages[1:]:
                     table.append(["", job.strip()])
             try:
                 progress = (self.progress / float(self.progress_total))
@@ -342,11 +382,11 @@ class IOManager:
 
     @contextmanager
     def job(self, job_text):
-        self.job_add(job_text)
+        job_id = self.job_add(job_text)
         try:
             yield
         finally:
-            self.job_del(job_text)
+            self.job_del(job_id)
 
     def job_wrapper(self, job_text):
         def outer_wrapper(wrapped_function):
@@ -357,9 +397,35 @@ class IOManager:
             return inner_wrapper
         return outer_wrapper
 
-    def _clear_last_job(self):
+    def _clear_last_job(self, flush=True):
         if self._status_line_present and TTY:
-            write_to_stream(STDOUT_WRITER, "\r\033[K")
+            # Some terminals respond very quickly to \e[K: They
+            # immediately clear the line and redraw their screen right
+            # away, which means an empty line is visible for a brief
+            # moment. It takes some (short amount of) time until we
+            # write the next line. This means we alternate between
+            # "empty line" and "line with text", which means:
+            # Flickering.
+            #
+            # We can avoid this to some degree by not flushing stdout
+            # after writing \e[K.
+            #
+            # We can only do this if we know for certain that we will
+            # write more data soon. For example, this is the case when
+            # _clear_last_job() is followed by _write_current_job().
+            #
+            # It is *not* possible to always omit the flush. Sometimes,
+            # _clear_last_job() is called before _write(..., err=True),
+            # meaning the first call writes to stdout and the second one
+            # to stderr. This results in garbled output because the \e[K
+            # is still lingering in the buffer for stdout.
+            #
+            # Note to future readers: If this causes issues and it's too
+            # hard to understand, then simply remove the 'flush' flag
+            # and go back to always flushing the buffers. This is mostly
+            # a cosmetic issue and it's better to have some flickering
+            # than completely garbled output.
+            write_to_stream(STDOUT_WRITER, "\r\033[K", flush=flush)
             self._status_line_present = False
 
     def _signal_handler_thread_body(self):
@@ -367,7 +433,7 @@ class IOManager:
             self.progress_show()
             if not self._waiting_for_input:  # do not block and ignore SIGINT while .ask()ing
                 with self.lock:
-                    self._clear_last_job()
+                    self._clear_last_job(flush=False)
                     self._write_current_job()
             if QUIT_EVENT.is_set():
                 if SHUTDOWN_EVENT_HARD.wait(0.1):
@@ -411,21 +477,29 @@ class IOManager:
         self._write_current_job()
 
     def _write_current_job(self):
-        if self.jobs and TTY:
-            line = "{} ".format(blue(self._spinner_character()))
-            try:
-                progress = (self.progress / float(self.progress_total))
-            except ZeroDivisionError:
-                pass
+        current_job = self.jobs.current_job
+        if TTY:
+            if current_job:
+                line = "{} ".format(blue(self._spinner_character()))
+                try:
+                    progress = (self.progress / float(self.progress_total))
+                except ZeroDivisionError:
+                    pass
+                else:
+                    progress_text = "{:.1f}%  ".format(progress * 100)
+                    line += bold(progress_text)
+                line += current_job
+                write_to_stream(
+                    STDOUT_WRITER,
+                    trim_visible_len_to(line, get_terminal_size().columns),
+                )
+                self._status_line_present = True
             else:
-                progress_text = "{:.1f}%  ".format(progress * 100)
-                line += bold(progress_text)
-            line += self.jobs[-1]
-            write_to_stream(
-                STDOUT_WRITER,
-                trim_visible_len_to(line, get_terminal_size().columns),
-            )
-            self._status_line_present = True
+                # If there are jobs to be displayed, the call above to
+                # write_to_stream() will flush stdout. But if there are
+                # none, then a \e[K might still be lingering in the
+                # buffer. See comment in _clear_last_job().
+                STDOUT_WRITER.flush()
 
 
 io = IOManager()
